# -*- coding: utf-8 -*-

import os
import re

from collections import defaultdict

######################################################################
# Constants
######################################################################

_ENCODING = 'utf8'

#: Positive infinity (for similarity functions)
_INF = 1e300

# Part of speech constants
ADJ, ADJ_SAT, ADV, NOUN, VERB = 'a', 's', 'r', 'n', 'v'
POS_LIST = [NOUN, VERB, ADJ, ADJ_SAT, ADV]
_FILEMAP = {ADJ: 'adj', ADV: 'adv', NOUN: 'noun', VERB: 'verb'}
_pos_numbers = {NOUN: 1, VERB: 2, ADJ: 3, ADV: 4, ADJ_SAT: 5}
_pos_names = dict(tup[::-1] for tup in _pos_numbers.items())
_synset_types = {1: NOUN, 2: VERB, 3: ADJ, 4: ADV, 5: ADJ_SAT}

# Max depth by WN version, simulate_root, and POS.
WN_MAX_DEPTH = {'3.0': {False: {'n': 19, 'v': 12, 'a': 0, 'r': 0},
                        True:  {'n': 20, 'v': 13, 'a': 1, 'r': 1}}
                }

#: A list of file identifiers for all the fileids used by wordnet.
_FILES = (
    'cntlist.rev',
    'lexnames',
    'index.sense',
    'index.adj',
    'index.adv',
    'index.noun',
    'index.verb',
    'data.adj',
    'data.adv',
    'data.noun',
    'data.verb',
    'adj.exc',
    'adv.exc',
    'noun.exc',
    'verb.exc',
)

MORPHOLOGICAL_SUBSTITUTIONS = {
    NOUN: [
        ('s', ''),
        ('ses', 's'),
        ('ves', 'f'),
        ('xes', 'x'),
        ('zes', 'z'),
        ('ches', 'ch'),
        ('shes', 'sh'),
        ('men', 'man'),
        ('ies', 'y'),
    ],
    VERB: [
        ('s', ''),
        ('ies', 'y'),
        ('es', 'e'),
        ('es', ''),
        ('ed', 'e'),
        ('ed', ''),
        ('ing', 'e'),
        ('ing', ''),
    ],
    ADJ: [('er', ''), ('est', ''), ('er', 'e'), ('est', 'e')],
    ADV: [],
}

MORPHOLOGICAL_SUBSTITUTIONS[ADJ_SAT] = MORPHOLOGICAL_SUBSTITUTIONS[ADJ]


# A table of strings that are used to express verb frames.
VERB_FRAME_STRINGS = (
    None,
    "Something %s",
    "Somebody %s",
    "It is %sing",
    "Something is %sing PP",
    "Something %s something Adjective/Noun",
    "Something %s Adjective/Noun",
    "Somebody %s Adjective",
    "Somebody %s something",
    "Somebody %s somebody",
    "Something %s somebody",
    "Something %s something",
    "Something %s to somebody",
    "Somebody %s on something",
    "Somebody %s somebody something",
    "Somebody %s something to somebody",
    "Somebody %s something from somebody",
    "Somebody %s somebody with something",
    "Somebody %s somebody of something",
    "Somebody %s something on somebody",
    "Somebody %s somebody PP",
    "Somebody %s something PP",
    "Somebody %s PP",
    "Somebody's (body part) %s",
    "Somebody %s somebody to INFINITIVE",
    "Somebody %s somebody INFINITIVE",
    "Somebody %s that CLAUSE",
    "Somebody %s to somebody",
    "Somebody %s to INFINITIVE",
    "Somebody %s whether INFINITIVE",
    "Somebody %s somebody into V-ing something",
    "Somebody %s something with something",
    "Somebody %s INFINITIVE",
    "Somebody %s VERB-ing",
    "It %s that CLAUSE",
    "Something %s INFINITIVE",
)

SENSENUM_RE = re.compile(r'\.[\d]+\.')


def load_exception_map():
    # load the exception file data into memory
    exception_map = {}
    for pos, suffix in _FILEMAP.items():
        exception_map[pos] = {}
        with open(wordnet_dir+'%s.exc' % suffix) as fin:
            for line in fin:
                terms = line.strip().split()
                exception_map[pos][terms[0]] = terms[1:]
    exception_map[ADJ_SAT] = exception_map[ADJ]
    return exception_map


def load_lemma_pos_offset_map():
    """ Not use in this library, kept for reference to NLTK. """
    lemma_pos_offset_map = defaultdict(dict)
    ##pos_lemma_offset_map = defaultdict(dict)
    for suffix in _FILEMAP.values():
        # parse each line of the file (ignoring comment lines)
        with open(wordnet_dir+'index.%s' % suffix) as fin:
            for i, line in enumerate(fin):
                if line.startswith(' '):
                    continue
                _iter = iter(line.split())
                def _next_token():
                    return next(_iter)
                try:
                    # get the lemma and part-of-speech
                    lemma = _next_token()
                    pos = _next_token()
                    # get the number of synsets for this lemma
                    n_synsets = int(_next_token())
                    assert n_synsets > 0
                    # get and ignore the pointer symbols for all synsets of
                    # this lemma
                    n_pointers = int(_next_token())
                    [_next_token() for _ in range(n_pointers)]
                    # same as number of synsets
                    n_senses = int(_next_token())
                    assert n_synsets == n_senses
                    # get and ignore number of senses ranked according to
                    # frequency
                    _next_token()
                    # get synset offsets
                    synset_offsets = [int(_next_token()) for _ in range(n_synsets)]

                # raise more informative error with file name and line number
                except (AssertionError, ValueError) as e:
                    tup = ('index.%s' % suffix), (i + 1), e
                    raise WordNetError('file %s, line %i: %s' % tup)

                # map lemmas and parts of speech to synsets
                lemma_pos_offset_map[lemma][pos] = synset_offsets
                ##pos_lemma_offset_map[pos][lemma] = synset_offsets
                if pos == ADJ:
                    lemma_pos_offset_map[lemma][ADJ_SAT] = synset_offsets
                    ##pos_lemma_offset_map[ADJ_SAT][lemma] = synset_offsets
    return lemma_pos_offset_map##, pos_lemma_offset_map

def load_lexnames():
    lexnames = []
    with open(wordnet_dir+'lexnames') as fin:
        # Load the lexnames
        for i, line in enumerate(fin):
            index, lexname, _ = line.split()
            assert int(index) == i
            lexnames.append(lexname)
    return lexnames


wn_data_dir = os.path.dirname(os.path.abspath(__file__)) + '/data/'
wordnet_dir = os.path.dirname(os.path.abspath(__file__)) + '/data/wordnet/'
wordnet_ic_dir = os.path.dirname(os.path.abspath(__file__)) + '/data/wordnet_ic/'
omw_dir = os.path.dirname(os.path.abspath(__file__)) + '/data/omw/'
exception_map = load_exception_map()
##lemma_pos_offset_map = load_lemma_pos_offset_map()
lexnames = load_lexnames()

__all__ = [
'_ENCODING',
'_INF',
'POS_LIST',
'ADJ', 'ADJ_SAT', 'ADV', 'NOUN', 'VERB',
'_FILEMAP',
'_pos_numbers',
'_pos_names',
'_synset_types',
'_FILES',
'MORPHOLOGICAL_SUBSTITUTIONS',
'VERB_FRAME_STRINGS',
'SENSENUM_RE',
'wn_data_dir',
'wordnet_dir', 'wordnet_ic_dir', 'omw_dir',
'exception_map', 'lexnames',
'WN_MAX_DEPTH']



# -*- coding: utf-8 -*-

from collections import deque

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

class WordNetError(Exception):
    """An exception class for wordnet-related errors."""

class FakeSynset:
    def __init__(self, name):
        self._name = '*ROOT*'
        self._min_depth = self._max_depth = 0
        self.hypernyms = lambda: []
        self.instance_hypernyms = lambda: []

    def min_depth(self):
        return self._min_depth

    def max_depth(self):
        return self._max_depth

    def __repr__(self):
        return "%s('%s')" % (type(self).__name__, self._name)

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other):
        return self._name == other._name

    def __ne__(self, other):
        return self._name != other._name

    def __lt__(self, other):
        return self._name < other._name

class WordNetObject(object):
    """A common base class for lemmas and synsets."""

    def hypernyms(self):
        return self._related('@')

    def _hypernyms(self):
        return self._related('@')

    def instance_hypernyms(self):
        return self._related('@i')

    def _instance_hypernyms(self):
        return self._related('@i')

    def hyponyms(self):
        return self._related('~')

    def instance_hyponyms(self):
        return self._related('~i')

    def member_holonyms(self):
        return self._related('#m')

    def substance_holonyms(self):
        return self._related('#s')

    def part_holonyms(self):
        return self._related('#p')

    def member_meronyms(self):
        return self._related('%m')

    def substance_meronyms(self):
        return self._related('%s')

    def part_meronyms(self):
        return self._related('%p')

    def topic_domains(self):
        return self._related(';c')

    def in_topic_domains(self):
        return self._related('-c')

    def region_domains(self):
        return self._related(';r')

    def in_region_domains(self):
        return self._related('-r')

    def usage_domains(self):
        return self._related(';u')

    def in_usage_domains(self):
        return self._related('-u')

    def attributes(self):
        return self._related('=')

    def entailments(self):
        return self._related('*')

    def causes(self):
        return self._related('>')

    def also_sees(self):
        return self._related('^')

    def verb_groups(self):
        return self._related('$')

    def similar_tos(self):
        return self._related('&')

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other):
        return self._name == other._name

    def __ne__(self, other):
        return self._name != other._name

    def __lt__(self, other):
        return self._name < other._name


def breadth_first(tree, children=iter, maxdepth=-1):
    """Traverse the nodes of a tree in breadth-first order.
    (No need to check for cycles.)
    The first argument should be the tree root;
    children should be a function taking as argument a tree node
    and returning an iterator of the node's children.
    """
    queue = deque([(tree, 0)])

    while queue:
        node, depth = queue.popleft()
        yield node

        if depth != maxdepth:
            try:
                queue.extend((c, depth + 1) for c in children(node))
            except TypeError:
                pass


def per_chunk(iterable, n=1, fillvalue=None):
    """
    From http://stackoverflow.com/a/8991553/610569
        >>> list(per_chunk('abcdefghi', n=2))
        [('a', 'b'), ('c', 'd'), ('e', 'f'), ('g', 'h'), ('i', None)]
        >>> list(per_chunk('abcdefghi', n=3))
        [('a', 'b', 'c'), ('d', 'e', 'f'), ('g', 'h', 'i')]
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)



# -*- coding: utf-8 -*-

from wn.constants import _pos_numbers
from wn.utils import WordNetObject

class Lemma(WordNetObject):
    def __init__(self, name, lexname_index, lex_id, syntactic_marker,
                 synset_offset=None, synset_pos=None,
                 synset_name=None, lemma_pointers=None, lang='eng'):
        self._name = name
        self._syntactic_marker = syntactic_marker
        self._lexname_index = lexname_index
        self._lex_id = lex_id
        self._lang = lang
        self._synset_offset = synset_offset
        self._synset_pos = synset_pos
        self._synset_name = synset_name
        self._lemma_pointers = lemma_pointers
        ##self._frame_strings = []
        ##self._frame_ids = []

    def synset(self):
        if hasattr(self, '_synset'):
            return self._synset
        else: # Fetch the synset from `_synset_offset_cache`
            self._synset = _synset_offset_cache[self._synset_pos][self._synset_offset]
        return self._synset

    def key(self):
        """
        This function can only be used after wordnet has been initialized after
        `_synset_offset_cache` has been populated and lemma can access the
        synsets' pos.

        From NLTK:
            # set sense keys for Lemma objects - note that this has to be
            # done afterwards so that the relations are available
        """
        if hasattr(self, '_key'):
            return self._key
        if self._synset_pos == 's':
            ss = _synset_offset_cache[self._synset_pos][self._synset_offset]
            head_lemma = ss.similar_tos()[0]._lemmas[0]
            head_name = head_lemma._name
            head_id = '%02d' % int(head_lemma._lex_id)
        else:
            head_name = head_id = ''
        sense_key_tuple = (self._name, _pos_numbers[self._synset_pos],
                           int(self._lexname_index), int(self._lex_id),
                           head_name, head_id)
        self._key = ('%s%%%d:%02d:%02d:%s:%s' % sense_key_tuple).lower()
        return self._key

    def count(self):
        if hasattr(self, '_count'):
            return self._count
        if self._lang != 'eng' and self.key() in _lemmakey_to_count:
            self._count = 0
        else:
            self._count = _lemmakey_to_count[self.key()]
        return self._count

    def name(self):
        return self._name

    def syntactic_marker(self):
        return self._syntactic_marker

    def frame_strings(self):
        return self._frame_strings

    def frame_ids(self):
        return self._frame_ids

    def lang(self):
        return self._lang

    def __repr__(self):
        tup = type(self).__name__, self._synset_name, self._name
        return "%s('%s.%s')" % tup

    def antonyms(self):
        return self._related('!')

    def derivationally_related_forms(self):
        return self._related('+')

    def pertainyms(self):
        return self._related('\\')

    def _related(self, relation_symbol):
        if (self._name, relation_symbol) not in self._lemma_pointers:
            return []
        return [
            _synset_offset_cache[pos][offset]._lemmas[lemma_index]
            for pos, offset, lemma_index in self._lemma_pointers[
                self._name, relation_symbol
            ]
        ]



# -*- coding: utf-8 -*-

import math
from collections import deque

from wn.utils import FakeSynset, WordNetError
from wn.constants import WN_MAX_DEPTH

def find_root_hypernyms(synset):
    """
    Note: Not used anywhere in this library, keeping it for reference to NLTK.
    Get the topmost hypernyms of this synset in WordNet.
    """
    result = []
    seen = set()
    todo = [synset]
    max_depth, min_depth
    while todo:
        next_synset = todo.pop()
        if next_synset not in seen:
            seen.add(next_synset)
            next_hypernyms = (
                next_synset.hypernyms() + next_synset.instance_hypernyms()
            )
            if not next_hypernyms:
                result.append(next_synset)
            else:
                todo.extend(next_hypernyms)
    return result

def find_shortest_hypernym_paths_to_root(synset, simulate_root=False):
    if synset._name == '*ROOT*':
        return {synset: 0}

    queue = deque([(synset, 0)])
    path = {}

    while queue:
        s, depth = queue.popleft()
        if s in path:
            continue
        path[s] = depth
        depth += 1
        queue.extend((hyp, depth) for hyp in s._hypernyms())
        queue.extend((hyp, depth) for hyp in s._instance_hypernyms())

    if simulate_root:
        fake_synset = FakeSynset(None)
        fake_synset._name = '*ROOT*'
        path[fake_synset] = max(path.values()) + 1

    return path


class WordNetPaths:
    def common_hypernyms(self, synset1, synset2):
        """
        Find all synsets that are hypernyms of this synset and the
        other synset.
        :return: The synsets that are hypernyms of both synsets.
        """
        return list(synset1.hypernyms_set().intersection(synset2.hypernyms_set()))

    def lowest_common_hypernyms(self, synset1, synset2,
                                simulate_root=False, use_min_depth=False):
        """
        Get a list of lowest synset(s) that both synsets have as a hypernym.
        When `use_min_depth == False` this means that the synset which appears
        as a hypernym of both `self` and `other` with the lowest maximum depth
        is returned or if there are multiple such synsets at the same depth
        they are all returned
        However, if `use_min_depth == True` then the synset(s) which has/have
        the lowest minimum depth and appear(s) in both paths is/are returned.
        By setting the use_min_depth flag to True, the behavior of NLTK2 can be
        preserved. This was changed in NLTK3 to give more accurate results in a
        small set of cases, generally with synsets concerning people. (eg:
        'chef.n.01', 'fireman.n.01', etc.)
        This method is an implementation of Ted Pedersen's "Lowest Common
        Subsumer" method from the Perl Wordnet module. It can return either
        "self" or "other" if they are a hypernym of the other.
        :type other: Synset
        :param other: other input synset
        :type simulate_root: bool
        :param simulate_root: The various verb taxonomies do not
            share a single root which disallows this metric from working for
            synsets that are not connected. This flag (False by default)
            creates a fake root that connects all the taxonomies. Set it
            to True to enable this behavior. For the noun taxonomy,
            there is usually a default root except for WordNet version 1.6.
            If you are using wordnet 1.6, a fake root will need to be added
            for nouns as well.
        :type use_min_depth: bool
        :param use_min_depth: This setting mimics older (v2) behavior of NLTK
            wordnet If True, will use the min_depth function to calculate the
            lowest common hypernyms. This is known to give strange results for
            some synset pairs (eg: 'chef.n.01', 'fireman.n.01') but is retained
            for backwards compatibility
        :return: The synsets that are the lowest common hypernyms of both
            synsets
        """
        synsets = self.common_hypernyms(synset1, synset2)
        if simulate_root:
            fake_synset = FakeSynset(None)
            fake_synset._name = '*ROOT*'
            fake_synset.hypernyms = lambda: []
            fake_synset.instance_hypernyms = lambda: []
            synsets.append(fake_synset)

        try:
            if use_min_depth:
                max_depth = max(s.min_depth() for s in synsets)
                unsorted_lch = [s for s in synsets if s.min_depth() == max_depth]
            else:
                max_depth = max(s.max_depth() for s in synsets)
                unsorted_lch = [s for s in synsets if s.max_depth() == max_depth]
            return sorted(unsorted_lch)
        except ValueError:
            return []

    def shortest_path_distance(self, synset1, synset2, simulate_root=False):
        """
        Returns the distance of the shortest path linking the two synsets (if
        one exists). For each synset, all the ancestor nodes and their
        distances are recorded and compared. The ancestor node common to both
        synsets that can be reached with the minimum number of traversals is
        used. If no ancestor nodes are common, None is returned. If a node is
        compared with itself 0 is returned.
        :type other: Synset
        :param other: The Synset to which the shortest path will be found.
        :return: The number of edges in the shortest path connecting the two
        nodes, or None if no path exists.
        """
        if synset1 == synset2:
            return 0
        # Find the shortest hypernym path to *ROOT*
        dist_dict1 = find_shortest_hypernym_paths_to_root(synset1, simulate_root)
        dist_dict2 = find_shortest_hypernym_paths_to_root(synset2, simulate_root)
        # For each ancestor synset common to both subject synsets, find the
        # connecting path length. Return the shortest of these.
        inf = float('inf')
        path_distance = inf
        for synset, d1 in dist_dict1.items():
            d2 = dist_dict2.get(synset, inf)
            path_distance = min(path_distance, d1 + d2)
        return None if math.isinf(path_distance) else path_distance

    def path_similarity(self, synset1, synset2, verbose=False, simulate_root=True,
                        if_none_return=None):
        """
        Path Distance Similarity:
        Return a score denoting how similar two word senses are, based on the
        shortest path that connects the senses in the is-a (hypernym/hypnoym)
        taxonomy. The score is in the range 0 to 1, except in those cases where
        a path cannot be found (will only be true for verbs as there are many
        distinct verb taxonomies), in which case None is returned. A score of
        1 represents identity i.e. comparing a sense with itself will return 1.
        :type other: Synset
        :param other: The ``Synset`` that this ``Synset`` is being compared to.
        :type simulate_root: bool
        :param simulate_root: The various verb taxonomies do not
        share a single root which disallows this metric from working for
        synsets that are not connected. This flag (True by default)
        creates a fake root that connects all the taxonomies. Set it
        to false to disable this behavior. For the noun taxonomy,
        there is usually a default root except for WordNet version 1.6.
        If you are using wordnet 1.6, a fake root will be added for nouns
        as well.
        :return: A score denoting the similarity of the two ``Synset`` objects,
        normally between 0 and 1. None is returned if no connecting path
        could be found. 1 is returned if a ``Synset`` is compared with
        itself.
        """
        distance = self.shortest_path_distance(synset1, synset2,
            simulate_root=simulate_root and synset1._needs_root())
        if distance is None or distance < 0:
            return if_none_return
        return 1.0 / (distance + 1)

    def _compute_max_depth(self, pos, simulate_root):
        """
        Compute the max depth for the given part of speech.  This is
        used by the lch similarity metric.
        """
        depth = 0
        for ii in self.all_synsets(pos):
            try:
                depth = max(depth, ii.max_depth())
            except RuntimeError:
                print(ii)
        if simulate_root:
            depth += 1
        self._max_depth[pos] = depth

    def lch_similarity(self, synset1, synset2,
                       simulate_root=True, _max_depth=None,
                       if_none_return=None):
        """
        Leacock Chodorow Similarity:
        Return a score denoting how similar two word senses are, based on the
        shortest path that connects the senses (as above) and the maximum depth
        of the taxonomy in which the senses occur. The relationship is given as
        -log(p/2d) where p is the shortest path length and d is the taxonomy
        depth.
        :type  other: Synset
        :param other: The ``Synset`` that this ``Synset`` is being compared to.
        :type simulate_root: bool
        :param simulate_root: The various verb taxonomies do not
            share a single root which disallows this metric from working for
            synsets that are not connected. This flag (True by default)
            creates a fake root that connects all the taxonomies. Set it
            to false to disable this behavior. For the noun taxonomy,
            there is usually a default root except for WordNet version 1.6.
            If you are using wordnet 1.6, a fake root will be added for nouns
            as well.
        :return: A score denoting the similarity of the two ``Synset`` objects,
            normally greater than 0. None is returned if no connecting path
            could be found. If a ``Synset`` is compared with itself, the
            maximum score is returned, which varies depending on the taxonomy
            depth.
        """

        if synset1._pos != synset2._pos:
            raise WordNetError(
                'Computing the lch similarity requires '
                '%s and %s to have the same part of speech.' % (synset1, synset2)
            )

        # Note the s`imlulate_root` here only tries to simulate
        # Ultimately the synset's need_root and simulate_root decides
        # whether root is really needed
        need_root = synset1._needs_root() and simulate_root
        # Hack to handle adjective and adverbs where _needs_root() returns None.
        if need_root == None:
            if if_none_return:
                need_root = True
            else: # Emulate NLTK's behavior to return None.
                return if_none_return
        # FIXME: how to make subclass overwrite values in kwargs?
        # By default use the static value from wn.constants
        depth = _max_depth if _max_depth else WN_MAX_DEPTH['3.0'][need_root][synset1._pos]
        distance = self.shortest_path_distance(synset1, synset2,
                        simulate_root=need_root)

        if distance is None or distance < 0 or depth == 0:
            return if_none_return
        return -math.log((distance + 1) / (2.0 * depth))

    def wup_similarity(self, synset1, synset2,
                       verbose=False, simulate_root=True,
                       if_none_return=None):
        """
        Wu-Palmer Similarity:
        Return a score denoting how similar two word senses are, based on the
        depth of the two senses in the taxonomy and that of their Least Common
        Subsumer (most specific ancestor node). Previously, the scores computed
        by this implementation did _not_ always agree with those given by
        Pedersen's Perl implementation of WordNet Similarity. However, with
        the addition of the simulate_root flag (see below), the score for
        verbs now almost always agree but not always for nouns.
        The LCS does not necessarily feature in the shortest path connecting
        the two senses, as it is by definition the common ancestor deepest in
        the taxonomy, not closest to the two senses. Typically, however, it
        will so feature. Where multiple candidates for the LCS exist, that
        whose shortest path to the root node is the longest will be selected.
        Where the LCS has multiple paths to the root, the longer path is used
        for the purposes of the calculation.
        :type  other: Synset
        :param other: The ``Synset`` that this ``Synset`` is being compared to.
        :type simulate_root: bool
        :param simulate_root: The various verb taxonomies do not
        share a single root which disallows this metric from working for
        synsets that are not connected. This flag (True by default)
        creates a fake root that connects all the taxonomies. Set it
        to false to disable this behavior. For the noun taxonomy,
        there is usually a default root except for WordNet version 1.6.
        If you are using wordnet 1.6, a fake root will be added for nouns
        as well.
        :return: A float score denoting the similarity of the two ``Synset``
        objects, normally greater than zero. If no connecting path between
        the two senses can be found, None is returned.
        """

        # Note the s`imlulate_root` here only tries to simulate
        # Ultimately the synset's need_root and simulate_root decides
        # whether root is really needed
        need_root = synset1._needs_root() and simulate_root
        # Note that to preserve behavior from NLTK2 we set use_min_depth=True
        # It is possible that more accurate results could be obtained by
        # removing this setting and it should be tested later on
        subsumers = self.lowest_common_hypernyms(synset1, synset2,
                        simulate_root=need_root, use_min_depth=True)

        # If no LCS was found return None
        if len(subsumers) == 0:
            return if_none_return

        subsumer = synset1 if synset1 in subsumers else subsumers[0]

        # Get the longest path from the LCS to the root,
        # including a correction:
        # - add one because the calculations include both the start and end
        #   nodes
        depth = subsumer.max_depth() + 1

        # Note: No need for an additional add-one correction for non-nouns
        # to account for an imaginary root node because that is now
        # automatically handled by simulate_root
        # if subsumer._pos != NOUN:
        #     depth += 1

        # Get the shortest path from the LCS to each of the synsets it is
        # subsuming.  Add this to the LCS path length to get the path
        # length from each synset to the root.
        len1 = self.shortest_path_distance(synset1, subsumer,
                simulate_root=simulate_root and need_root)
        len2 = self.shortest_path_distance(synset2, subsumer,
                simulate_root=simulate_root and need_root)
        if len1 is None or len2 is None:
            return if_none_return
        len1 += depth
        len2 += depth
        return (2.0 * depth) / (len1 + len2)



# -*- coding: utf-8 -*-

import math
from collections import defaultdict

from wn.constants import wordnet_ic_dir
from wn.reader import parse_wordnet_ic_line
from wn.utils import WordNetError


class WordNetInformationContent:
    def __init__(self, corpus, resnik=False, add1=False):
        self.available_corpora = ['bnc', 'brown', 'semcor',
                                  'semcorraw', 'shak', 'treebank']
        if corpus not in self.available_corpora:
            msg = '{} not in list of availabel Wordnet IC files: {}'.format(self.available_corpora)
            raise WordNetError(msg)

        self.corpus = corpus
        self.with_resnik = '-resnik' if resnik else ''
        self.with_add1 = '-add1' if add1 else ''
        self.ic_filename = '/ic-{}{}{}.dat'.format(self.corpus,
                                                self.with_resnik,
                                                self.with_add1)

        self.ic = defaultdict(dict)
        with open(wordnet_ic_dir + self.ic_filename) as fin:
            next(fin) # Skip the first line.
            for line in fin:
                offset, value, pos, has_root = parse_wordnet_ic_line(line)
                if has_root:
                    self.ic[pos][0] = self.ic[pos].get(0, 0) + value
                if value != 0:
                    self.ic[pos][offset] = value


class InformationContentSimilarities:
    def information_content(self, synset, ic):
        if synset._pos in ic.ic and ic.ic[synset._pos]:
            icpos = ic.ic[synset._pos]
        else:
            msg = 'Information content file has no entries for part-of-speech: %s'
            raise WordNetError(msg % synset._pos)
        counts = icpos[synset._offset]
        if counts == 0:
            return _INF
        else:
            return -math.log(counts / icpos[0])

    def _lcs_ic(self, synset1, synset2, ic):
        """
        Get the information content of the least common subsumer that has
        the highest information content value.  If two nodes have no
        explicit common subsumer, assume that they share an artificial
        root node that is the hypernym of all explicit roots.
        :type synset1: Synset
        :param synset1: First input synset.
        :type synset2: Synset
        :param synset2: Second input synset.  Must be the same part of
        speech as the first synset.
        :type  ic: dict
        :param ic: an information content object (as returned by ``load_ic()``).
        :return: The information content of the two synsets and their most
        informative subsumer
        """
        if synset1._pos != synset2._pos:
            raise WordNetError(
                'Computing the least common subsumer requires '
                '%s and %s to have the same part of speech.' % (synset1, synset2)
            )

        ic1 = self.information_content(synset1, ic)
        ic2 = self.information_content(synset2, ic)
        # Find common hypernyms of synset1 and synset2
        subsumers = list(synset1.hypernyms_set().intersection(synset2.hypernyms_set()))
        if len(subsumers) == 0:
            subsumer_ic = 0
        else:
            subsumer_ic = max(self.information_content(s, ic) for s in subsumers)

        return ic1, ic2, subsumer_ic

    def res_similarity(self, synset1, synset2, ic):
        """
        Resnik Similarity:
        Return a score denoting how similar two word senses are, based on the
        Information Content (IC) of the Least Common Subsumer (most specific
        ancestor node).
        :type  other: Synset
        :param other: The ``Synset`` that this ``Synset`` is being compared to.
        :type ic: dict
        :param ic: an information content object (as returned by
        ``nltk.corpus.wordnet_ic.ic()``).
        :return: A float score denoting the similarity of the two ``Synset``
        objects. Synsets whose LCS is the root node of the taxonomy will
        have a score of 0 (e.g. N['dog'][0] and N['table'][0]).
        """
        ic1, ic2, lcs_ic = self._lcs_ic(synset1, synset2, ic)
        return lcs_ic

    def jcn_similarity(self, synset1, synset2, ic):
        """
        Jiang-Conrath Similarity:
        Return a score denoting how similar two word senses are, based on the
        Information Content (IC) of the Least Common Subsumer (most specific
        ancestor node) and that of the two input Synsets. The relationship is
        given by the equation 1 / (IC(s1) + IC(s2) - 2 * IC(lcs)).
        :type  other: Synset
        :param other: The ``Synset`` that this ``Synset`` is being compared to.
        :type  ic: dict
        :param ic: an information content object (as returned by
        ``nltk.corpus.wordnet_ic.ic()``).
        :return: A float score denoting the similarity of the two ``Synset``
        objects.
        """

        if synset1 == synset2:
            return _INF

        ic1, ic2, lcs_ic = self._lcs_ic(synset1, synset2, ic)

        # If either of the input synsets are the root synset, or have a
        # frequency of 0 (sparse data problem), return 0.
        if ic1 == 0 or ic2 == 0:
            return 0

        ic_difference = ic1 + ic2 - 2 * lcs_ic

        if ic_difference == 0:
            return _INF

        return 1 / ic_difference

    def lin_similarity(self, synset1, synset2, ic):
        """
        Lin Similarity:
        Return a score denoting how similar two word senses are, based on the
        Information Content (IC) of the Least Common Subsumer (most specific
        ancestor node) and that of the two input Synsets. The relationship is
        given by the equation 2 * IC(lcs) / (IC(s1) + IC(s2)).
        :type other: Synset
        :param other: The ``Synset`` that this ``Synset`` is being compared to.
        :type ic: dict
        :param ic: an information content object (as returned by
        ``nltk.corpus.wordnet_ic.ic()``).
        :return: A float score denoting the similarity of the two ``Synset``
        objects, in the range 0 to 1.
        """

        ic1, ic2, lcs_ic = self._lcs_ic(synset1, synset2, ic)
        return (2.0 * lcs_ic) / (ic1 + ic2)



# -*- coding: utf-8 -*-

import math
from itertools import chain
from operator import itemgetter
from collections import defaultdict

import wn.path
from wn.constants import *
from wn.lemma import Lemma
from wn.morphy import morphy
from wn.omw import OpenMultilingualWordNet as OMW
from wn.utils import WordNetObject, WordNetError
from wn.utils import breadth_first, FakeSynset

class Synset(WordNetObject):
    def __init__(self, offset, pos, name, lexname_index, lexname,
                 definition, examples=None, pointers=None,
                 lemmas=None, wordnet_line=None):

        self._offset = offset
        self._pos = pos
        self._name = name
        self._lexname = lexname  # lexicographer name.

        self._definition = definition
        self._examples = examples if examples else []

        self._pointers = pointers if pointers else defaultdict(set)
        self._lemmas = lemmas

        self._wordnet_line = wordnet_line

    def __repr__(self):
        return "%s('%s')" % (type(self).__name__, self._name)

    def offset(self):
        return self._offset

    def pos(self):
        return self._pos

    def name(self):
        return self._name

    def definition(self):
        return self._definition

    def examples(self):
        return self._examples

    def lexname(self):
        return self._lexname

    def _needs_root(self): # Assumes Wordnet >=V3.0.
        if self._pos == NOUN:
            return False
        elif self._pos == VERB:
            return True

    def lemmas(self, lang='eng'):
        '''Return all the lemma objects associated with the synset'''
        if lang == 'eng':
            return self._lemmas
        else:
            if lang not in _lang_to_lemmas_to_offsets.keys():
                # If not in cache, temporarily create an OMW object.
                # Get the mappings from the staticmethod `custom_lemmas()`
                offsets_to_lemmas, lemmas_to_offsets = OMW.custom_lemmas(lang)
                # Load it up.
                _lang_to_offsets_to_lemma[lang] = offsets_to_lemmas
                _lang_to_lemmas_to_offsets[lang] = lemmas_to_offsets
            # Return the list of lemmas from the OMW cache.
            lemmark = []
            for lemma_name in self.lemma_names(lang):
                _name = lemma_name
                _lexname_index = _lex_id = 0
                _syntactic_marker = None
                _synset_name = self._name
                _synset_offset = self._offset
                _synset_pos = self._pos
                lemma = Lemma(_name, _lexname_index, _lex_id, _syntactic_marker,
                              _synset_offset, _synset_pos, _synset_name,
                              lemma_pointers=None, lang=lang)
                lemmark.append(lemma)
            return lemmark

    def lemma_names(self, lang='eng'):
        if lang == 'eng':
            return [l._name for l in self._lemmas]
        else:
            if lang not in _lang_to_lemmas_to_offsets.keys():
                # If not in cache, temporarily create an OMW object.
                # Get the mappings from the staticmethod `custom_lemmas()`
                offsets_to_lemmas, lemmas_to_offsets = OMW.custom_lemmas(lang)
                # Load it up.
                _lang_to_offsets_to_lemma[lang] = offsets_to_lemmas
                _lang_to_lemmas_to_offsets[lang] = lemmas_to_offsets

            # Return the list of lemmas from the OMW cache.
            if self._pos in _lang_to_offsets_to_lemma[lang]:
                return _lang_to_offsets_to_lemma[lang][self._pos][self._offset]
            elif self._pos == 's' and 'a' in _lang_to_offsets_to_lemma[lang]:
                return _lang_to_offsets_to_lemma[lang]['a'][self._offset]
            else:
                return []

    def _related(self, relation_symbol, sort=True):
        if relation_symbol not in self._pointers:
            return []
        related_synsets = []
        for pos, offset in self._pointers[relation_symbol]:
            if pos in ['s', 'a']:
                try:
                    related_synset = _synset_offset_cache['a'][offset]
                except:
                    try:
                        related_synset = _synset_offset_cache['s'][offset]
                    except:
                        raise WordNetError('Part-of-Speech and Offset combination not found in WordNet: {} + {}'.format(pos, offset))
            else:
                related_synset = _synset_offset_cache[pos][offset]
            related_synsets.append(related_synset)

        return sorted(related_synsets) if sort else related_synsets

    def _hypernym_paths(self):
        """
        Get the path(s) from this synset to the root, where each path is a
        list of the synset nodes traversed on the way to the root.
        :return: A list of lists, where each list gives the node sequence
        connecting the initial ``Synset`` node and a root node.
        """
        paths = []
        hypernyms = self.hypernyms() + self.instance_hypernyms()
        if len(hypernyms) == 0:
            paths  = [[self]]
        for hypernym in hypernyms:
            for ancestor_list in hypernym._hypernym_paths():
                ancestor_list.append(self)
                paths.append(ancestor_list)
        return paths

    def _init_hypernym_paths(self):
        """
        Note: This can only be done after the whole wordnet is read, so
        this will be called on the fly when user tries to access:
        (i) _hyperpaths, (ii) _min_depth, (iii) _max_depth or (iv) _root_hypernyms
        """
        self._hyperpaths = self._hypernym_paths()
        # Compute the path related statistics.
        if self._hyperpaths:
            self._min_depth = min(len(path) for path in self._hyperpaths) - 1
            self._max_depth = max(len(path) for path in self._hyperpaths) - 1
        else:
            self._min_depth = self._max_depth = 0
        # Compute the store the root hypernyms.
        self._root_hypernyms = list(set([path[0] for path in self._hyperpaths]))
        # Initialize the hypernyms_set for `common_hypernyms()`
        self._hypernyms_set = set(chain(*self._hyperpaths))
        self._hypernyms_set.remove(self)

    def hypernym_paths(self):
        if not hasattr(self, '_hyperpaths'):
            self._init_hypernym_paths()
        return self._hyperpaths

    def min_depth(self):
        if not hasattr(self, '_min_depth'):
            self._init_hypernym_paths()
        return self._min_depth

    def max_depth(self):
        if not hasattr(self, '_max_depth'):
            self._init_hypernym_paths()
        return self._max_depth

    def root_hypernyms(self):
        if not hasattr(self, '_root_hypernyms'):
            self._init_hypernym_paths()
        return self._root_hypernyms

    def hypernyms_set(self):
        if not hasattr(self, '_hypernyms_set'):
            self._init_hypernym_paths()
        return self._hypernyms_set

    def closure(self, rel, depth=-1):
        """Return the transitive closure of source under the rel
        relationship, breadth-first
            >>> from nltk.corpus import wordnet as wn
            >>> dog = wn.synset('dog.n.01')
            >>> hyp = lambda s:s.hypernyms()
            >>> list(dog.closure(hyp))
            [Synset('canine.n.02'), Synset('domestic_animal.n.01'),
            Synset('carnivore.n.01'), Synset('animal.n.01'),
            Synset('placental.n.01'), Synset('organism.n.01'),
            Synset('mammal.n.01'), Synset('living_thing.n.01'),
            Synset('vertebrate.n.01'), Synset('whole.n.02'),
            Synset('chordate.n.01'), Synset('object.n.01'),
            Synset('physical_entity.n.01'), Synset('entity.n.01')]
        """
        synset_offsets = []
        for synset in breadth_first(self, rel, depth):
            if synset._offset != self._offset:
                if synset._offset not in synset_offsets:
                    synset_offsets.append(synset._offset)
                    yield synset

    def tree(self, rel, depth=-1, cut_mark=None):
        """
        >>> from nltk.corpus import wordnet as wn
        >>> dog = wn.synset('dog.n.01')
        >>> hyp = lambda s:s.hypernyms()
        >>> from pprint import pprint
        >>> pprint(dog.tree(hyp))
        [Synset('dog.n.01'),
         [Synset('canine.n.02'),
          [Synset('carnivore.n.01'),
           [Synset('placental.n.01'),
            [Synset('mammal.n.01'),
             [Synset('vertebrate.n.01'),
              [Synset('chordate.n.01'),
               [Synset('animal.n.01'),
                [Synset('organism.n.01'),
                 [Synset('living_thing.n.01'),
                  [Synset('whole.n.02'),
                   [Synset('object.n.01'),
                    [Synset('physical_entity.n.01'),
                     [Synset('entity.n.01')]]]]]]]]]]]]],
         [Synset('domestic_animal.n.01'),
          [Synset('animal.n.01'),
           [Synset('organism.n.01'),
            [Synset('living_thing.n.01'),
             [Synset('whole.n.02'),
              [Synset('object.n.01'),
               [Synset('physical_entity.n.01'), [Synset('entity.n.01')]]]]]]]]]
        """
        tree = [self]
        if depth != 0:
            tree += [x.tree(rel, depth - 1, cut_mark) for x in rel(self)]
        elif cut_mark:
            tree += [cut_mark]
        return tree

    def hypernym_distances(self, distance=0, simulate_root=False):
        """
        Get the path(s) from this synset to the root, counting the distance
        of each node from the initial node on the way. A set of
        (synset, distance) tuples is returned.
        :type distance: int
        :param distance: the distance (number of edges) from this hypernym to
        the original hypernym ``Synset`` on which this method was called.
        :return: A set of ``(Synset, int)`` tuples where each ``Synset`` is
        a hypernym of the first ``Synset``.
        """
        distances = set([(self, distance)])
        for hypernym in self._hypernyms() + self._instance_hypernyms():
            distances |= hypernym.hypernym_distances(distance + 1, simulate_root=False)
        if simulate_root:
            fake_synset = FakeSynset(None)
            fake_synset._name = '*ROOT*'
            fake_synset_distance = max(distances, key=itemgetter(1))[1]
            distances.add((fake_synset, fake_synset_distance + 1))
        return distances



# !/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import re
from itertools import chain
from collections import defaultdict

from wn.constants import *
from wn.info import InformationContentSimilarities
from wn.path import WordNetPaths
from wn.morphy import morphy
from wn.omw import OpenMultilingualWordNet
from wn.reader import fix_inconsistent_line
from wn.reader import parse_wordnet_line
from wn.reader import parse_index_line
from wn.reader import parse_lemma_pos_index
from wn.reader import parse_sense_key
from wn.utils import WordNetError, FakeSynset


# Abusing builtins here but this is the only way I can think of.
# A index that provides the file offset

# Map from lemma -> pos -> synset_index -> offset
__builtins__['_lemma_pos_offset_map'] = defaultdict(dict)
# A cache so we don't have to reconstuct synsets
# Map from pos -> offset -> synset
__builtins__['_synset_offset_cache'] = defaultdict(dict)

# A cache to store the wordnet data of multiple languages
__builtins__['_lang_to_offsets_to_lemma'] = defaultdict(dict)
__builtins__['_lang_to_lemmas_to_offsets'] = defaultdict(dict)

# Map from sensekey -> count
__builtins__['_lemmakey_to_count'] = {}

__version__ = '0.0.18'

class WordNet(WordNetPaths, InformationContentSimilarities, OpenMultilingualWordNet):
    def __init__(self, wordnet_data_dir=wordnet_dir, lexname_type=None, wordnet_33=False):
        self.wordnet_data_dir = wordnet_data_dir
        self.lexname_type = lexname_type
        self.wordnet_33 = wordnet_33
        # Initializes the `_lemma_pos_offset_map` and `_pos_lemma_offset_map`
        # from wn.constants.
        self._load_lemma_pos_offset_map()
        # Initializes the `_synset_offset_cache`
        # from wn.constants.
        self._load_all_synsets()
        # Initialize all lemma's count.
        self._load_all_lemma_counts()

        self._synset_offset_cache = _synset_offset_cache
        self._lemma_pos_offset_map = _lemma_pos_offset_map

        self._lang_to_offsets_to_lemma = _lang_to_offsets_to_lemma
        self._lang_to_lemmas_to_offsets = _lang_to_lemmas_to_offsets

        self._lemmakey_to_count = _lemmakey_to_count

    def _load_lemma_pos_offset_map(self):
        for pos_tag in _FILEMAP.values():
            filename = os.path.join(self.wordnet_data_dir, 'index.{}'.format(pos_tag))
            with io.open(filename, encoding='utf8') as fin:
                for line in fin:
                    if line.startswith(' '):
                        continue
                    try:
                        lemma, pos, synset_offsets = parse_index_line(line)
                    except: # When there's inconsistencies.
                        if self.wordnet_33:
                            lemma, pos, synset_offsets = parse_index_line(fix_inconsistent_line(line))
                        else:
                            raise WordNetError('Error parsing:\n{}\nfrom {}'.format(line, filename))
                    # Cache the map.
                    _lemma_pos_offset_map[lemma][pos] = synset_offsets
                    if pos == 'a':
                        _lemma_pos_offset_map[lemma]['s'] = synset_offsets

    def _load_all_synsets(self):
        for pos_tag in _FILEMAP.values():
            filename = os.path.join(self.wordnet_data_dir, 'data.{}'.format(pos_tag))
            with io.open(filename, encoding='utf8') as fin:
                for line in fin:
                    # Skip documentation and empty lines.
                    if line.startswith(' ') or not line.strip():
                        continue
                    try:
                        synset, lemmas = parse_wordnet_line(line, lexname_type=self.lexname_type)
                        _synset_offset_cache[synset._pos][synset._offset] = synset
                    except:
                        err_msg = "Error parsing this line from {}:\n".format('data.{}'.format(pos_tag))
                        raise WordNetError(err_msg + line)

    def _load_all_lemma_counts(self):
        filename = os.path.join(self.wordnet_data_dir, 'cntlist.rev')
        with open(filename) as fin:
            for line in fin:
                lemma_key, _, count = line.strip().split()
                _lemmakey_to_count[lemma_key] = int(count)

    def synset_from_pos_and_offset(self, pos, offset):
        assert pos in POS_LIST, WordNetError('Part-of-Speech should be one of this: {}'.format(POS_LIST))
        offset = int(offset)
        try:
            return _synset_offset_cache[pos][offset]
        except:
            if pos == 's' and offset in _synset_offset_cache['a']:
                return _synset_offset_cache['a'][offset]
            if pos == 'a' and offset in _synset_offset_cache['s']:
                return _synset_offset_cache['s'][offset]
            raise WordNetError('Part-of-Speech and Offset combination not found in WordNet: {} + {}'.format(pos, offset))

    def synset(self, lemma_pos_index):
        # Parse the lemma_pos_index string.
        pos, offset = parse_lemma_pos_index(lemma_pos_index)
        # load synset information from the appropriate file
        synset = self.synset_from_pos_and_offset(pos, offset)
        # Return the synset object.
        return synset

    def synsets(self, lemma, pos=None, lang='eng', check_exceptions=True):
        """
        Load all synsets with a given lemma and part of speech tag.
        If no pos is specified, all synsets for all parts of speech
        will be loaded.
        If lang is specified, all the synsets associated with the lemma name
        of that language will be returned.
        """
        lemma = lemma.lower()
        pos_tags = POS_LIST if pos == None else [pos]
        if lang == 'eng':
            list_of_synsets = []
            for p in pos_tags:
                form = morphy(lemma, p, check_exceptions)
                for offset in _lemma_pos_offset_map[form].get(p, []):
                    if offset in _synset_offset_cache[p]:
                        list_of_synsets.append(_synset_offset_cache[p][offset])
                    else:
                        if pos == 's' and offset in _synset_offset_cache['a']:
                            list_of_synsets.append(_synset_offset_cache['a'][offset])
                        elif pos == 'a' and offset in _synset_offset_cache['s']:
                            list_of_synsets.append(_synset_offset_cache['s'][offset])

            return list_of_synsets
        else:
            # Tries to cache the OMW for the first time if not used before.
            self._load_lang_data(lang)
            # Iterate through the _lang_to_lemmas_to_offsets to get the offsets.
            list_of_offsets = []
            for p in pos_tags:
                if p == 's': # Skips the 's' tag.
                    continue
                if lemma in _lang_to_lemmas_to_offsets[lang][p]:
                    for offset in _lang_to_lemmas_to_offsets[lang][p][lemma]:
                        list_of_offsets.append((p, offset))
            return [self.synset_from_pos_and_offset(p, offset)
                    for p, offset in set(list_of_offsets)]

    def synset_from_sense_key(self, sense_key):
        lemma, pos, lex_id = parse_sense_key(sense_key)
        synset_id = '.'.join([lemma, pos, lex_id])
        return self.synset(synset_id)

    def all_synsets(self, pos=None):
        """Iterate over all synsets with a given part of speech tag.
        If no pos is specified, all synsets for all parts of speech
        will be loaded.
        """
        pos_tags = POS_LIST if pos is None else [pos]

        for _pos in pos_tags:
            for offset, ss in _synset_offset_cache[_pos].items():
                yield ss

    def all_lemma_names(self, pos=None, lang='eng'):
        if lang == 'eng':
            for lemma_name in _lemma_pos_offset_map:
                if pos in _lemma_pos_offset_map[lemma_name] or pos == None:
                    yield lemma_name
        else:
            # Tries to cache the OMW for the first time if not used before.
            self._load_lang_data(lang)
            for lemma_name in _lang_to_lemmas_to_offsets[lang][pos]:
                yield lemma_name

    def words(self, lang='lang'):
        """return lemmas of the given language as list of words"""
        return self.all_lemma_names(lang=lang)

    def lemma(self, name, lang='eng'):
        '''Return lemma object that matches the name'''
        # cannot simply split on first '.',
        # e.g.: '.45_caliber.a.01..45_caliber'
        separator = SENSENUM_RE.search(name).end()
        synset_name, lemma_name = name[: separator - 1], name[separator:]

        synset = self.synset(synset_name)
        for lemma in synset.lemmas(lang):
            if lemma._name == lemma_name:
                return lemma
        raise WordNetError('no lemma %r in %r' % (lemma_name, synset_name))

    def _compute_max_depth_once(self, pos, simulate_root):
        """
        Compute the max depth for the given part of speech.  This is
        used by the lch similarity metric.

        This function should never be used!!!
        It should be computed once, then put into wn.constants.
        """
        depth = 0
        for ss in self.all_synsets(pos):
            try:
                depth = max(depth, ss.max_depth())
            except RuntimeError:
                _msg = '{} throws error when searching for max_depth'.format(ss)
                raise WordNetError(_msg)
        if simulate_root:
            depth += 1
        return depth

    def get_version(self):
        filename = self.wordnet_data_dir+'/data.adj'
        with io.open(filename, encoding='utf8') as fin:
            for line in fin:
                match = re.search(r'WordNet (\d+\.\d+) Copyright', line)
                if match:
                    self._version = match.group(1)
                    return self._version
        raise WordNetError("Cannot find version number in {}".format(filename))

    def version(self):
        if hasattr(self, '_version'):
            return self._version
        return self.get_version()

    def _compute_max_depth(self, pos, simulate_root):
        """
        Compute the max depth for the given part of speech.  This is
        used by the lch similarity metric.
        """
        pos = 'a' if pos == 's' else pos
        # Try to fetch wordnet's max_depth from constants.
        version = self.version()
        if version in WN_MAX_DEPTH or hasattr(self, '_max_depth'):
            self._max_depth = WN_MAX_DEPTH
            return self._max_depth[version][simulate_root][pos]

        # Compute the _wn_max_depth for this wordnet version for the first time.
        self._max_depth = {version: {True: {}, False: {}}}
        for _pos in POS_LIST:
            if _pos == 's': # Skips the satellite adjective.
                continue
            depth = self._max_depth[version][False][_pos] = 0
            for ss in self.all_synsets(_pos):
                depth = max(depth, ss.max_depth())
            self._max_depth[version][True][_pos] = depth + 1
            self._max_depth[version][False][_pos] = depth
        return self._max_depth[version][simulate_root][pos]



#-*- coding: utf-8 -*-

#############################################################
# Morphy: adapted from Oliver Steele's pywordnet
#############################################################

from itertools import chain, islice

from wn.constants import *

def morphy(form, pos=None, check_exceptions=True):
    """
    Find a possible base form for the given form, with the given
    part of speech, by checking WordNet's list of exceptional
    forms, and by recursively stripping affixes for this part of
    speech until a form in WordNet is found.
    """

    if pos is None:
        morphy = _morphy
        analyses = chain(a for p in POS_LIST for a in morphy(form, p))
    else:
        analyses = _morphy(form, pos, check_exceptions)

    # get the first one we find
    first = list(islice(analyses, 1))
    if len(first) == 1:
        return first[0]
    else:
        return None


def _morphy(form, pos, check_exceptions=True):
    # from jordanbg:
    # Given an original string x
    # 1. Apply rules once to the input to get y1, y2, y3, etc.
    # 2. Return all that are in the database
    # 3. If there are no matches, keep applying rules until you either
    #    find a match or you can't go any further

    exceptions = exception_map[pos]
    substitutions = MORPHOLOGICAL_SUBSTITUTIONS[pos]

    def apply_rules(forms):
        return [
            form[: -len(old)] + new
            for form in forms
            for old, new in substitutions
            if form.endswith(old)
        ]

    def filter_forms(forms):
        result = []
        seen = set()
        for form in forms:
            if form in _lemma_pos_offset_map:
                if pos in _lemma_pos_offset_map[form]:
                    if form not in seen:
                        result.append(form)
                        seen.add(form)
        return result

    # 0. Check the exception lists
    if check_exceptions:
        if form in exceptions:
            return filter_forms([form] + exceptions[form])

    # 1. Apply rules once to the input to get y1, y2, y3, etc.
    forms = apply_rules([form])

    # 2. Return all that are in the database (and check the original too)
    results = filter_forms([form] + forms)
    if results:
        return results

    # 3. If there are no matches, keep applying rules until we find a match
    while forms:
        forms = apply_rules(forms)
        results = filter_forms(forms)
        if results:
            return results

    # Return an empty list if we can't find anything
    return []
